/**
 * Abstraction for threading to allow porting.
 * Currently using either pthread or C++ 11 threads.
 */

#ifndef PLS_THREAD_H
#define PLS_THREAD_H

#include <functional>

// platform specific includes
#include "pls/internal/base/choose_threading.h"
#ifdef PLS_USING_PTHREADS
    #include <pthread.h>
#include <iostream>

#elif PLS_USING_CPP_THREADS
    #include <thread>
#endif

namespace pls {
    namespace internal {
        namespace base {
            using thread_entrypoint = void();

            class this_thread {
                template<typename Function, typename State>
                friend class thread;
#ifdef PLS_USING_PTHREADS
                static pthread_key_t local_storage_key_;
                static bool local_storage_key_initialized_;
#endif
#ifdef PLS_USING_CPP_THREADS
                static thread_local void* local_storage_;
#endif

            public:
                static void yield() {
#ifdef PLS_USING_PTHREADS
                    pthread_yield();
#endif
#ifdef PLS_USING_CPP_THREADS
                    std::this_thread::yield();
#endif
                }

                template<typename T>
                static T* state() {
#ifdef PLS_USING_PTHREADS
                    return reinterpret_cast<T*>(pthread_getspecific(local_storage_key_));
#endif
#ifdef PLS_USING_CPP_THREADS
                    reinterpret_cast<T*>(local_storage_);
#endif
                }

                template<typename T>
                static void set_state(const T& state) {
#ifdef PLS_USING_PTHREADS
                    *reinterpret_cast<T*>(pthread_getspecific(local_storage_key_)) = state;
#endif
#ifdef PLS_USING_CPP_THREADS
                    *reinterpret_cast<T*>(local_storage_) = state;
#endif
                }
            };

            template<typename Function, typename State>
            class thread {
                friend class this_thread;
                // Keep a copy of the function (lambda) in this object to make sure it is valid when called!
                Function function_;
                // Keep the local state we hold in here
                State state_;

                // Keep handle to native implementation
#ifdef PLS_USING_PTHREADS
                pthread_t pthread_thread_;
#endif
#ifdef PLS_USING_CPP_THREADS
                std::thread std_thread_;
#endif

#ifdef PLS_USING_PTHREADS
                static void* start_pthread_internal(void* thread_pointer) {
                    auto my_thread = reinterpret_cast<thread*>(thread_pointer);
                    pthread_setspecific(this_thread::local_storage_key_, (void*)&my_thread->state_);
                    my_thread->function_();
                    pthread_exit(nullptr);
                }

#endif

            public:
#ifdef PLS_USING_PTHREADS
                explicit thread(const Function& function, const State& state):
                        function_{function},
                        state_{state},
                        pthread_thread_{} {}

                void start() {
                    if (!this_thread::local_storage_key_initialized_) {
                        pthread_key_create(&this_thread::local_storage_key_, nullptr);
                        this_thread::local_storage_key_initialized_ = true;
                    }

                    pthread_create(&pthread_thread_, nullptr, start_pthread_internal, (void *)(this));
                }
#endif
#ifdef PLS_USING_CPP_THREADS
                explicit thread(const Function& function, const State& state):
                function_{function},
                state_{state},
                std_thread_{0} {};

                void start() {
                    std_thread_ = std::thread([=](){
                        local_storage = reinterpret_cast<void*><&this->state_>;
                        this->function_();
                    });
                }
#endif
            public:
                void join() {
#ifdef PLS_USING_PTHREADS
                    pthread_join(pthread_thread_, nullptr);
#endif
#ifdef PLS_USING_CPP_THREADS
                    std_thread_.join();
#endif
                }

                // make object move only
                thread(thread&&) noexcept = default;
                thread& operator=(thread&&) noexcept = default;

                thread(const thread&) = delete;
                thread& operator=(const thread&) = delete;
            };

            template<typename Function, typename State>
            thread<Function, State> create_thread(const Function& function, const State& state) {
                return thread<Function, State>(function, state);
            }

        }
    }

}

#endif //PLS_THREAD_H
